class MatchBase:
    
    def __init__(self, template: str) -> None:
        self.template = template

    def index(self, probe: str) -> int:
        """
        Exact matching, return the matched index.
        """
        raise Exception('Must be implemented in subclass!')

    def blurry_index(self, probe: str, allowed_mismatches: int):
        """
        Blurry matching, return the matched index.
        """
        raise Exception('Must be implemented in subclass!')

    def has(self, probe: str, allowed_mismatches=0) -> bool:
        if allowed_mismatches == 0:
            return self.index(probe) >= 0
        else:
            return self.blurry_index(probe, allowed_mismatches) >= 0


def center(msg: str, decorator='#', width=80):
    n = 2
    x = ' ' * int(width / 2 - n - len(msg) / 2)
    s0 = x + msg
    s1 = s0 + ' ' * (width - len(s0) - 4)
    s = '\n' + decorator * width + '\n'
    s += decorator * n + s1 + decorator * n + '\n'
    s += decorator * width
    print(s)
    return s


def clean_seq(fasta_text: str) -> str:
    clear_seq = ''
    seqs = fasta_text.split('\n')
    for line in seqs:
        if line.strip():
            if line.startswith('>'):
                continue
            else:
                clear_seq += line.strip()
    return clear_seq


import logging
from logging import ERROR, FATAL, StreamHandler, WARN, WARNING
from logging import FileHandler
from logging import Formatter
from datetime import datetime

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR

def getLogger(name, to_console=True, to_file=False, **kwargs):
    """
    `level`: str, default: `logging.INFO`
        
    If `to_console` is `True`:
        - `console_fmt`: str, default: `'%(asctime)s %(name)s.%(levelname)s # %(message)s'`
        - `console_datefmt`: str, default: `'%Y-%m-%d %H:%M:%S'`

    If `to_file` is `True`:
        - `output_dir`: str, default: `'./'`
        - `timestamp`: str, deafult: `'%Y%m%d%H%M%S'`, for log file name
        - `file_fmt`: str, default: `console_fmt` or `'%(asctime)s %(name)s.%(levelname)s # %(message)s'`
        - `file_datefmt`: str, default: `console_datefmt` or `'%Y-%m-%d %H:%M:%S'`
    """

    _level = kwargs.get('level', logging.INFO)  # 日志级别

    logger = logging.getLogger(name)
    logger.setLevel(_level)

    if to_console:
        stream_handler = StreamHandler()
        _console_fmt = kwargs.get(
            'console_fmt', '%(asctime)s %(name)s.%(levelname)s: %(message)s')
        _console_datefmt = kwargs.get('console_datefmt', r'%Y-%m-%d %H:%M:%S')
        formatter = Formatter(fmt=_console_fmt, datefmt=_console_datefmt)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    if to_file:
        _output_dir = kwargs.get('output_dir', './')
        _timestamp = kwargs.get('timestamp', r'%Y%m%d%H%M%S')
        if _timestamp is None:
            logfile = f'{_output_dir}/{name}.log'
        else:
            time_str = datetime.now().strftime(_timestamp)
            logfile = f'{_output_dir}/{name}.{time_str}.log'
        file_handler = FileHandler(logfile, 'w')
        _file_fmt = kwargs.get('file_fmt', None)
        if _file_fmt is None:
            _file_fmt = kwargs.get('console_fmt', None)
        if _file_fmt is None:
            _file_fmt = '%(asctime)s %(name)s.%(levelname)s: %(message)s'
        _file_datefmt = kwargs.get('file_datefmt')
        if _file_datefmt is None:
            _file_datefmt = kwargs.get('console_datefmt')
        if _file_datefmt is None:
            _file_datefmt = r'%Y-%m-%d %H:%M:%S'
        formatter = Formatter(fmt=_file_fmt, datefmt=_file_datefmt)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger



import re

def openFile(filepath: str, mode='r'):
    if filepath.endswith('.gz'):
        import gzip
        _mode = 'rt' if mode == 'r' else mode
        return gzip.open(filepath, _mode)
    if filepath.endswith('.bz2'):
        import bz2
        _mode = 'rt' if mode == 'r' else mode
        return bz2.open(filepath, _mode)
    else:
        return open(filepath, mode)


def write_once(fp: str, data):
    """
        写入一个数据到一个文件，然后关闭文件
        :param fp:
        :param data: 任意类型的数据
        :param data_type: 数据的类型，与Python3的类型注解一致
            支持的类型有:
            1. str 写入一个字符串
            2. List[str] 写入一个字符串数组
        :return:
        """

    text = None
    if isinstance(data, str):
        text = data
    elif isinstance(data, list):
        if any([isinstance(i, str) for i in data]):
            text = '\n'.join(data)
        elif any([isinstance(i, int) for i in data]):
            text = '\n'.join([int(i) for i in data])

    if text is None:
        raise ValueError('Not supported type')

    with open(fp, 'w') as writer:
        writer.write(text)


def read_chunk(fp: str, chunk_size=512):
    with openFile(fp) as reader:
        while True:
            chunk = reader.read(chunk_size)
            if chunk:
                yield chunk
            else:
                break


def iterate_fasta(fn, replace_degenerate=True, seq_header=False):

    degenerate_pattern = re.compile('[YRWSMKBDHV]')

    def process(f):
        curr_header = ''
        curr_seq = ''
        for line in f:
            line = line.rstrip()
            if len(line) == 0:
                # Skip the blank line
                continue
            if line.startswith('>'):
                # Yield the current sequence (if there is one) and reset the
                # sequence being read
                if len(curr_seq) > 0:
                    if seq_header:
                        yield curr_seq, curr_header
                    else:
                        yield curr_seq
                curr_header = line
                curr_seq = ''
            else:
                # Append the sequence
                if replace_degenerate:
                    line = degenerate_pattern.sub('N', line)
                curr_seq += line
        
        if len(curr_seq) > 0:
            if seq_header:
                yield curr_seq, curr_header
            else:
                yield curr_seq

    with openFile(fn) as reader:
        yield from process(reader)


"""将fasta序列数字化，保存到hdf5文件中"""
from collections import defaultdict
from typing import Callable
import h5py
import numpy as np
from Bio import SeqIO
from barwet.io import openFile
from barwet.log import getLogger
logger = getLogger(__name__)


# digitalize a str into np array.
def digitalize(seq_str: str, keys: str = 'NAGCT') -> np.ndarray:
    mappings = defaultdict(int)
    for i, v in enumerate(keys):
        mappings[v.upper()] = i
        mappings[v.lower()] = i
    return np.array([mappings[c] for c in seq_str], dtype='i1')


## write to fasta file.
def write_to_fasta(input_file: str, output_file: str):
    with openFile(input_file) as reader, \
    openFile(output_file, 'wb') as writer:
        for line_n in reader:
            line = line_n.strip()
            if line.startswith('>'):
                logger.info(f'now at {line}')
                writed_line = line_n
            elif not line:
                writed_line = line_n
            else:
                writed_line = digitalize(line) + '\n'
            writer.write(writed_line.encode('utf8'))


## write to hdf5 file.
def write_to_hdf5(input_file: str,
                  output_file: str,
                  key_fn: Callable[[str], str] = None):
    """
    Args:
        input_file: .fasta or .fna
        output_file: .h5
        key_fn: get dataset name from <SeqRecord>.id
    """
    if key_fn is None: key_fn = lambda x: x
    with openFile(input_file) as reader, \
    h5py.File(output_file, 'w') as h5:
        for record in SeqIO.parse(reader, 'fasta'):
            chr_id = key_fn(record.id)
            logger.info(f'Now at {chr_id}')
            digital_seq = digitalize(record.seq)
            h5.create_dataset(chr_id, data=digital_seq)


if __name__ == '__main__':
    input_file = '/home/barwe/data/barwe/pd/human/HumanGRCh38.fna.gz'
    # output_file = '/home/barwe/data/barwe/pd/human/HumanGRCh38.digit.fna.gz'
    output_file = '/home/barwe/data/barwe/pd/human/HumanGRCh38.digit.h5'
    key_fn = lambda s: s.split(' ')[0].split('|')[-1]
    write_to_hdf5(input_file, output_file, key_fn)



from .base import MatchBase
# from base import MatchBase


class SundayMatch(MatchBase):
    def index(self, probe: str) -> int:
        """Exact match"""
        # 模板字符串和探针字符串的长度
        template = self.template
        nt = len(template)
        np = len(probe)
        # 统计 probe 中所有字符出现的最大索引位置
        char_pos = {}
        for idx, char in enumerate(probe):
            char_pos[char] = idx

        index = 0  # 母串上与子串首字符比较的位置
        while index <= nt - np:
            is_matched = True
            for i, pi in enumerate(probe):
                # 子串第i个字符与母串第index+i个字符不匹配时
                if template[index + i] != pi:
                    is_matched = False
                    # 下一个移动位置超过母串长度时，匹配失败
                    if index + np >= nt:
                        return -1
                    # 可以继续移动子串 #
                    # 判断母串上下一个移动位置的字符是否在子串中
                    pos = char_pos.get(template[index + np])
                    if pos is None:
                        index += np + 1
                    else:
                        index += np - pos
                    break
            if is_matched:
                return index
        return -1

    def blurry_index(self, probe: str, allowed_mismatches: int) -> int:
        """Blurry match"""
        # 模板字符串和探针字符串的长度
        template = self.template
        nt = len(template)
        np = len(probe)
        # 统计 probe 中所有字符出现的最大索引位置
        char_pos = {}
        for idx, char in enumerate(probe):
            char_pos[char] = idx

        index = 0  # 母串上与子串首字符比较的位置
        while index <= nt - np:
            num_mismatches = 0  # 当前位置下的错配数
            for i, pi in enumerate(probe):
                # 子串第i个字符与母串第index+i个字符不匹配时
                if template[index + i] != pi:
                    num_mismatches += 1
                    # 当错配数超过阈值时考虑移动子串
                    if num_mismatches > allowed_mismatches:
                        # 下一个移动位置超过母串长度时，匹配失败
                        if index + np >= nt:
                            return -1
                        # 否则可以继续移动子串
                        pos = char_pos.get(template[index + np])
                        if pos is None:
                            index += np + 1
                        else:
                            index += np - pos
                        break
            if num_mismatches <= allowed_mismatches:
                return index
        return -1


if __name__ == '__main__':
    template = "this is a simple example"
    probe = 'example'
    print(SundayMatch(template).blurry_index(probe, 0))
    print(SundayMatch(template).blurry_index(probe, 1))
    print(SundayMatch(template).blurry_index(probe, 2))
    print(SundayMatch(template).blurry_index(probe, 3))


import h5py
from typing import List
import numba as nb
import numpy as np
from Bio import SeqIO
from timeit import timeit
from barwet.io import openFile
from barwet.log import getLogger

logger = getLogger(__name__)


## 从文件加载用于测试的探针列表
def load_demo_probes(fp: str) -> List[str]:
    probes = []
    with openFile(fp) as r:
        for _p in r:
            p = _p.strip()
            if p:
                probes.append(p)
    return probes


## 将探针数字化
from digitalize_fasta import digitalize


def digitalize_probes(probes: List[str]) -> np.ndarray:
    p_num = len(probes)
    p_len = len(probes[0])
    digital_probes = np.zeros(shape=(p_num, p_len), dtype='i1')
    for i, p in enumerate(probes):
        digital_probes[i, :] = digitalize(p)
    return digital_probes


## Load digital genome from fasta file.
def load_digital_genome_from_fasta(fp: str):
    with openFile(fp) as reader:
        records = []
        for record in SeqIO.parse(reader, 'fasta'):
            print(record.id)
            seq_digit = np.array([int(i) for i in record.seq], 'i1')
            records.append(seq_digit)
            break
    return records


## Load digital genome from hdf5 file.
## Return a Generator object.
def load_digital_genome_from_hdf5(fp: str):
    chrdata = {}
    with h5py.File(fp, 'r') as db:
        for chrname in db.keys():
            chrseq = np.array(db[chrname], dtype='i1')
            chrdata[chrname] = chrseq
            # yield chrname, chrseq
    return chrdata


@nb.njit(nogil=True)
def search_probe_in_chr(chrseq: np.ndarray, probe: np.ndarray,
                        mismatches: int) -> int:
    c_len = chrseq.size
    p_len = probe.size

    char_pos = {}
    for idx, char in enumerate(probe):
        char_pos[char] = idx

    index = 0  # 母串上与子串首字符比较的位置
    while index <= c_len - p_len:
        num_mismatches = 0  # 当前位置下的错配数
        for i, pi in enumerate(probe):
            # 子串第i个字符与母串第index+i个字符不匹配时
            if chrseq[index + i] != pi:
                num_mismatches += 1
                # 当错配数超过阈值时考虑移动子串
                if num_mismatches > mismatches:
                    # 下一个移动位置超过母串长度时，匹配失败
                    if index + p_len >= c_len:
                        return -1
                    # 否则可以继续移动子串
                    pos = char_pos.get(chrseq[index + p_len])
                    if pos is None:
                        index += p_len + 1
                    else:
                        index += p_len - pos
                    break
        if num_mismatches <= mismatches:
            return index
    return -1


@nb.njit(nogil=True)
def search_probes_in_chr(chrseq, probes, mismatches):
    """
    Args:
        chrseq: [C_LEN,]
        probes: [P_NUM, P_LEN]
    Returns:
        np.ndarray[P_NUM,] index of each probe located in chromosome.
    """
    p_num = probes.shape[0]
    # logger.info(f'chromosome length: {c_len}, probe length: {p_len}, probe counts: {p_num}')

    positions = np.empty((p_num, ), dtype='int64')
    for idx in nb.prange(p_num):
        positions[idx] = search_probe_in_chr(chrseq, probes[idx, :],
                                             mismatches)

    return positions


from multiprocessing import pool


def search_async(chrseq, probes, mismatches, processes):
    """将单个探针的在单个染色体上的搜索下发到子进程"""
    p_num = probes.shape[0]
    positions = np.empty((p_num, ), dtype='int64')
    results: List[pool.AsyncResult] = []
    p = pool.Pool(processes)
    for i in nb.prange(p_num):
        args = [chrseq, probes[i, :], mismatches]
        results.append(p.apply_async(search_probe_in_chr, args))
    p.close()
    p.join()
    positions = [r.get() for r in results]
    return np.array(positions, dtype='int64')


if __name__ == '__main__':

    genome = load_digital_genome_from_hdf5(
        '/home/barwe/data/barwe/pd/human/HumanGRCh38.digit.h5')
    probes = digitalize_probes(
        load_demo_probes('/home/barwe/data/barwe/pd/human/probes.txt.gz')[:20])

    for chrname, chrseq in genome.items():
        logger.info(f'Searching in chromosome: {chrname}')
        # print(timeit(lambda: search_probes_in_chr(chrseq, probes, 1), number=2)) # 17.14
        print(timeit(lambda: search_async(chrseq, probes, 1, 4), number=1))  #
        print(timeit(lambda: search_async(chrseq, probes, 1, 6), number=1))  #
        print(timeit(lambda: search_async(chrseq, probes, 1, 8), number=1))  #
        break


import os

if __name__ == '__main__':
    import log
    import jsn
else:
    from . import log
    from . import jsn

logger = log.getLogger(__name__, level=log.INFO)
# logger = log.getLogger(__name__, level=log.DEBUG)


class Glovar(object):
    ## 单态
    __shared_state = {}

    ## 单例
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, gvfile=None):
        """
        Args:
            gvfile: str, JSON file storing global variables.
        """
        super().__init__()
        ## 从单态字典中查询字段
        shared_file_path = self.__shared_state.get('gvfile')
        # 字段不存在时添加字段
        if shared_file_path is None:
            if gvfile is None:
                raise ValueError(f'`gvfile` cannot be None: {gvfile}')
            else:
                self.__shared_state['gvfile'] = gvfile
        # 字段存在时使用原字段值
        else:
            if gvfile is not None:
                logger.warn(f'Existed instance, `gvfile` will be ignored')
        ## 从单态字典更新对象属性字典
        self.__dict__ = self.__shared_state

    def _load(self):
        if os.path.exists(self.gvfile):
            d = jsn.load(self.gvfile)
            if isinstance(d, dict):
                return d
        return {}

    def get(self, k, default=None):
        d = self._load()
        return d.get(k, default)

    def set(self, k, v):
        d = self._load()
        d[k] = v
        jsn.dump(d, self.gvfile)

    def close(self):
        _ = os.system(f'rm -f {self.gvfile}')


def get_glovar(gvfile=None):
    return Glovar(gvfile)

def get(k):
    return Glovar().get(k)

if __name__ == '__main__':

    gv1 = Glovar('./test.json')
    gv1.set('name', 'barwe')
    gv2 = Glovar('a')
    print(gv2.get('name'))

    print(id(gv1), id(gv2))
    print(gv1.gvfile, gv2.gvfile)
    print(gv1.__dict__, gv2.__dict__)



import os
import json

from hashlib import md5
from random import shuffle, randint

from typing import Any, Callable

def md5(string: str):
    from hashlib import md5 as HASH_FN
    return HASH_FN(string.encode()).hexdigest()

def hash_string(string: str, fn: Callable[[str, ], str]=None) -> str:
    """Generate unique ID for a string.
    """
    if fn is None:
        return md5(string.encode()).hexdigest()
    else:
        return fn(string)


## 对字典键值对的唯一性进行哈希
def to_str(data: Any) -> str:
    if isinstance(data, dict):
        ks = sorted(list(data.keys()))
        d = {k: to_str(data[k]) for k in ks}
        return json.dumps(d)
    else:
        return json.dumps(data)

def hash_dict_content(data: dict) -> str:
    s = to_str(data)
    return hash_string(s)

def random_id(length=64, use: str='NAa', chars=''):
    x = {
        'N': '1234567890',
        'a': 'abcdefghijklmnopqrstuvwxyz',
        'A': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
        'S': '~`!@#$%^&*()_-+=[]{}\|;:\'",.<>/?'
    }
    y = {'N': 10, 'a': 26, 'A': 26, 'S': 33}

    keys = []
    for k in x:
        if k in use:
            keys.append(k)
            chars += x[k]

    chars = [i for i in chars]
    shuffle(chars)

    seq = ''
    for i in range(length):
        seq += chars[randint(0, len(chars)-1)]

    return seq




from typing import Iterable

def sort(_iterable: Iterable, key=None, reverse=False):
    return sorted(_iterable, key, reverse)

def argsort(_iterable: Iterable, key=None, reverse=False):
    _it = sorted(enumerate(_iterable), key=key, reverse=reverse)
    index = [iv[0] for iv in _it]
    return index

def to_batch(arr, num_batches):
    size = len(arr)
    each_size = int(size / num_batches)
    batches = []
    for s in range(num_batches):
        e = (s + 1) * each_size
        if e > size:
            e = size
        batches.append(arr[s:e])
    return batches

def flatten(arr):
    flatted_arr = []
    for sub in arr:
        if isinstance(sub, Iterable):
            flatted_arr.extend(flatten(sub))
        else:
            flatted_arr.append(sub)
    return flatted_arr